from builtins import str, bytes, object
import time

from WMCore.Database.CMSCouch import CouchServer, Database
from WMCore.Lexicon import splitCouchServiceURL, sanitizeURL


class RequestDBReader(object):
    def __init__(self, couchURL, couchapp="ReqMgr"):
        # set the connection for local couchDB call
        self._commonInit(couchURL, couchapp)

    def _commonInit(self, couchURL, couchapp):
        """
        setting up comon variables for inherited class.
        inherited class should call this in their init function
        """
        if isinstance(couchURL, Database):
            self.couchDB = couchURL
            self.couchURL = self.couchDB['host']
            self.dbName = self.couchDB.name
            self.couchServer = CouchServer(self.couchURL)
        else:
            # NOTE: starting in CouchDB 3.x, we need to provide the couch credentials in
            # order to be able to write to the database, thus a RequestDBWriter object
            if isinstance(self.__class__, RequestDBReader):
                couchURL = sanitizeURL(couchURL)['url']
            self.couchURL, self.dbName = splitCouchServiceURL(couchURL)
            self.couchServer = CouchServer(self.couchURL)
            self.couchDB = self.couchServer.connectDatabase(self.dbName, False)
        self.couchapp = couchapp
        self.defaultStale = {"stale": "update_after"}

    def setDefaultStaleOptions(self, options):
        if not options:
            options = {}
        if 'stale' not in options:
            options.update(self.defaultStale)
        return options

    def _setNoStale(self):
        """
        Use this only for the unittest
        """
        self.defaultStale = {}

    def _getCouchView(self, view, options, keys=None):
        keys = keys or []
        options = self.setDefaultStaleOptions(options)

        if keys and isinstance(keys, (str, bytes)):
            keys = [keys]
        return self.couchDB.loadView(self.couchapp, view, options, keys)

    def _filterCouchInfo(self, couchInfo):
        # remove the couch specific information
        for key in ['_rev', '_attachments']:
            if key in couchInfo:
                del couchInfo[key]
        return

    def _formatCouchData(self, data, key="id", detail=True, filterCouch=True, returnDict=False):
        result = {}
        for row in data['rows']:
            if 'error' in row:
                continue
            if "doc" in row:
                if filterCouch:
                    self._filterCouchInfo(row["doc"])
                result[row[key]] = row["doc"]
            else:
                result[row[key]] = row["value"]
        if detail or returnDict:
            return result
        else:
            return list(result)

    def _getRequestByName(self, requestName, detail):
        """
        Retrieves a request dictionary from CouchDB
        :param requestName: string with the request name
        :param detail: boolean with False value for retrieving only the
            workflow name, or True for retrieving all its description
        :return: a list with the request name. Or a dictionary with the
            request description if detail=true
        """
        # this returns a dictionary with the workflow description, or
        # an empty dictionary if nothing is found in CouchDB
        result = self.couchDB.getDoc(requestName)
        if not result:
            return dict()
        if detail:
            result.pop('_attachments', None)
            result.pop('_rev', None)
            return {result['RequestName']: result}
        return [result['RequestName']]

    def _getRequestByNames(self, requestNames, detail):
        """
        'status': list of the status
        """
        options = {}
        options["include_docs"] = detail
        result = self.couchDB.allDocs(options, requestNames)
        return result

    def _getRequestByStatus(self, statusList, detail, limit, skip):
        """
        'status': list of the status
        """
        options = {}
        options["include_docs"] = detail

        if limit != None:
            options["limit"] = limit
        if skip != None:
            options["skip"] = skip
        keys = statusList
        return self._getCouchView("bystatus", options, keys)

    def _getRequestByStatusAndStartTime(self, status, detail, startTime):
        timeNow = int(time.time())
        options = {}
        options["include_docs"] = detail
        options["startkey"] = [status, startTime]
        options["endkey"] = [status, timeNow]
        options["descending"] = False

        return self._getCouchView("bystatusandtime", options)

    def _getRequestByStatusAndEndTime(self, status, detail, endTime):
        """
        'status': is the status of the workflow
        'endTime': unix timestamp for end time
        """
        options = {}
        options["include_docs"] = detail
        options["startkey"] = [status, 0]
        options["endkey"] = [status, endTime]
        options["descending"] = False

        return self._getCouchView("bystatusandtime", options)

    def _getRequestByTeamAndStatus(self, team, status, limit):
        """
        'status': is the status of the workflow
        'startTime': unix timestamp for start time
        """
        options = {}
        if limit:
            options["limit"] = limit
        if team and status:
            options["key"] = [team, status]
        elif team and not status:
            options["startkey"] = [team]
            options["endkey"] = [team, status]  # status = {}

        return self._getCouchView("byteamandstatus", options)

    def _getAllDocsByIDs(self, ids, include_docs=True):
        """
        keys is [id, ....]
        returns document
        """
        if len(ids) == 0:
            return []
        options = {}
        options["include_docs"] = include_docs
        result = self.couchDB.allDocs(options, ids)

        return result

    def getDBInstance(self):
        return self.couchDB

    def getRequestByNames(self, requestNames, detail=True):
        if len(requestNames) == 0:
            return {}
        if isinstance(requestNames, list) and len(requestNames) == 1:
            requestNames = requestNames[0]

        if isinstance(requestNames, (str, bytes)):
            requestInfo = self._getRequestByName(requestNames, detail=detail)
        else:
            requestInfo = self._getRequestByNames(requestNames, detail=detail)
            requestInfo = self._formatCouchData(requestInfo, detail=detail)
        return requestInfo

    def getRequestByStatus(self, statusList, detail=False, limit=None, skip=None):

        data = self._getRequestByStatus(statusList, detail, limit, skip)
        requestInfo = self._formatCouchData(data, detail=detail)

        return requestInfo

    def getRequestByStatusAndStartTime(self, status, detail=False, startTime=0):
        """
        Query for requests that are in a specific status since startTime.
        :param status: string with the workflow status
        :param detail: boolean flag used to return doc content or not
        :param startTime: unix start timestamp for your query
        :return: a list of request names
        """
        if startTime == 0:
            data = self._getRequestByStatus([status], detail, limit=None, skip=None)
        else:
            data = self._getRequestByStatusAndStartTime(status, detail, startTime)

        requestInfo = self._formatCouchData(data, detail=detail)
        return requestInfo

    def getRequestByStatusAndEndTime(self, status, detail=False, endTime=0):
        """
        Query for requests that are in a specific status until endTime.
        :param status: string with the workflow status
        :param detail: boolean flag used to return doc content or not
        :param endTime: unix end timestamp for your query
        :return: a list of request names
        """
        if endTime == 0:
            data = self._getRequestByStatus([status], detail, limit=None, skip=None)
        else:
            data = self._getRequestByStatusAndEndTime(status, detail, endTime)

        requestInfo = self._formatCouchData(data, detail=detail)
        return requestInfo

    def getRequestByTeamAndStatus(self, team, status, detail=False, limit=None):
        """
        'team': team name in which the workflow was assigned to.
        'status': a single status string.
        """
        if team and status:
            data = self._getRequestByTeamAndStatus(team, status, limit)
        elif team and not status:
            data = self._getRequestByTeamAndStatus(team, status={}, limit=limit)
        elif not team and not status:
            data = self._getRequestByTeamAndStatus(team={}, status={}, limit=limit)
        else:
            # nothing we can do with status only
            return

        requestInfo = self._formatCouchData(data, detail=detail)
        return requestInfo

    def getRequestByCouchView(self, view, options, keys=None, returnDict=True):
        keys = keys or []
        options.setdefault("include_docs", True)
        data = self._getCouchView(view, options, keys)
        requestInfo = self._formatCouchData(data, returnDict=returnDict)
        return requestInfo

    def getStatusAndTypeByRequest(self, requestNames):
        if isinstance(requestNames, (str, bytes)):
            requestNames = [requestNames]
        if len(requestNames) == 0:
            return {}
        data = self._getCouchView("byrequest", {}, requestNames)
        requestInfo = self._formatCouchData(data, returnDict=True)
        return requestInfo

    def getStepChainDatasetParentageByStatus(self, status):
        options = {}
        options["key"] = [False, status]
        data = self._getCouchView("byparentageflag", options)
        datasetParentageInfo = self._formatCouchData(data, returnDict=True)
        return datasetParentageInfo



from builtins import map, str, bytes
from future.utils import viewitems

import cherrypy
import hashlib
import hmac
import re

from Utils.Utilities import lowerCmsHeaders, encodeUnicodeToBytes
from Utils.PythonVersion import PY3

def get_user_info():
    "Helper function to return user based information of the request"
    return cherrypy.request.user


def user_info_from_headers(key, verbose=False):
    """Read the user information HTTP request headers added by front-end.
    Validates the HMAC on them to check for tampering, and if all is ok,
    returns user info object with the data from the headers."""
    # Set initial user information for this request
    log = cherrypy.log
    headers = lowerCmsHeaders(cherrypy.request.headers)
    user = {'dn': None, 'method': None, 'login': None, 'name': None, 'roles': {}}

    # Reject if request was not authenticated.
    if 'cms-auth-status' not in headers:
        if headers.get('Remote-Addr', '') == '127.0.0.1':
            # since we are using localhost access we can skip authentication
            log("DEBUG: non-authenticated request performed on localhost")
            return
        log("ERROR: authz denied, front-end headers not present")
        raise cherrypy.HTTPError(403, "You are not allowed to access this resource.")

    # If authentication is optional and wasn't done, accept.
    if headers['cms-auth-status'] == 'NONE':
        if verbose: log("DEBUG: authn optional and missing")
        return

    # Extract user information from the headers. Collect data required
    # for HMAC validation while processing headers.
    prefix = suffix = ""
    hkeys = sorted(headers.keys())
    for hk in hkeys:
        hk = hk.lower()
        if hk[0:9] in ("cms-authn", "cms-authz") and hk != "cms-authn-hmac":
            prefix += "h%xv%x" % (len(hk), len(headers[hk]))
            suffix += "%s%s" % (hk, headers[hk])
            hkname = hk.split('-', 2)[-1]
            if hk.startswith("cms-authn"):
                val = headers[hk]
                if hk in ("cms-authn-name", "cms-authn-dn"):
                    val = str(val) if PY3 else str(val, "utf-8")
                user[hkname] = val
            if hk.startswith("cms-authz"):
                user['roles'][hkname] = {'site': set(), 'group': set()}
                for r in headers[hk].split():
                    site_or_group, name = r.split(':')
                    user['roles'][hkname][site_or_group].add(name)

    # Check HMAC over authn/z headers with server key. If differs, reject.
    msg = prefix + "#" + suffix
    if PY3:
        key = encodeUnicodeToBytes(key)
        msg = encodeUnicodeToBytes(msg)
    cksum = hmac.new(key, msg, hashlib.sha1).hexdigest()
    if cksum != headers["cms-authn-hmac"]:
        log("ERROR: authz hmac mismatch, %s vs. %s" % (cksum, headers["cms-authn-hmac"]))
        raise cherrypy.HTTPError(403, "You are not allowed to access this resource.")

    # Authn/z is legal, accept
    if verbose:
        log("DEBUG: authn accepted for user %s" % user)
    return user


def authz_canonical(val):
    """Make a name canonical."""
    return re.sub(r"[^a-z0-9]+", "-", val.lower())


def authz_match(role=None, group=None, site=None, verbose=False):
    """Match user against authorisation requirements."""
    role = role or []
    group = group or []
    site = site or []
    user = get_user_info()
    log = cherrypy.log

    # If role, group or site are strings, convert to list first.
    role = (role and isinstance(role, (str, bytes)) and [role]) or role
    group = (group and isinstance(group, (str, bytes)) and [group]) or group
    site = (site and isinstance(site, (str, bytes)) and [site]) or site

    # Reformat all items into canonical format.
    role = role and list(map(authz_canonical, role))
    group = group and list(map(authz_canonical, group))
    site = site and list(map(authz_canonical, site))

    # If role, group and site are all empty, no authz requirements: pass
    if not (role or group or site):
        if verbose:
            log("DEBUG: authz accepted nil requirements for user %s" % user)
        return

    # Otherwise determine set intersection of requirements.
    for r, authz in viewitems((user and user['roles']) or {}):
        if (not role) or (r in role):
            if not (group or site):
                if verbose:
                    log("DEBUG: authz accepted role '%s' for user %s" % (r, user))
                return
            if set(group) & authz['group']:
                if verbose:
                    log("DEBUG: authz accepted role '%s' group %s for user %s" % (r, group, user))
                return
            if set(site) & authz['site']:
                if verbose:
                    log("DEBUG: authz accepted role '%s' site %s for user %s" % (r, site, user))
                return

    # Deny access, requirements weren't fulfilled
    log("ERROR: authz denied role %s group %s site %s for user %s" % (role, group, site, user))
    raise cherrypy.HTTPError(403, "You are not allowed to access this resource.")


def authz_user(role=None, group=None, site=None, key=None, verbose=False):
    """Default authorisation implementation: returns True for an authorised
    user, one with any of the requested roles for the given sites or groups.

    If no roles are specified, belonging to any of the requested sites or
    groups is sufficient."""
    role = role or []
    group = group or []
    site = site or []
    # Get user information from request headers, then run real matcher.
    assert getattr(cherrypy.request, "user", None) == None
    cherrypy.request.user = user_info_from_headers(key, verbose)
    authz_match(role, group, site, verbose)


_fake_warned = False


def authz_fake(role=None, group=None, site=None, verbose=False):
    """Fake authorisation routine."""
    role = role or []
    group = group or []
    site = site or []
    log = cherrypy.log
    global _fake_warned
    if not _fake_warned:
        if cherrypy.server.environment == 'production':
            log("ERROR: authz faking denied in production server")
            raise Exception("authz faking denied in production server")
        if cherrypy.server.socket_host not in ("127.0.0.1", "::1"):
            log("ERROR: authz faking denied on non-local interface")
            raise Exception("authz faking denied on non-local interface")
        log("WARNING: authz faking activated, only test use is permitted")
        _fake_warned = True
    cherrypy.request.user = {
        'method': 'fake',
        'login': 'fake_insecure_auth',
        'name': 'Fake Insecure Auth',
        'dn': '/CN=Fake Insecure Auth',
        'roles': {'fake': {'site': set(['fake']), 'group': set(['fake'])}}
    }

    if verbose:
        log("DEBUG: authz accepted bypass role %s group %s site %s for user %s"
            % (role, group, site, cherrypy.request.user))


class RESTAuth(cherrypy.Tool):
    """Restrict access based on CMSWEB authn/z information."""
    _key = None

    def __init__(self):
        cherrypy.Tool.__init__(self, 'before_request_body', authz_user, priority=60)

    def _setup(self):
        """Hook this tool into cherrypy request."""
        log = cherrypy.log
        hooks = cherrypy.request.hooks
        conf = self._merged_args()

        # Install authz policy. Load front-end hmac key on first use.
        prio = conf.pop("priority", self._priority)
        policy = conf.pop("policy", "user")
        if policy == "user":
            key_file = conf.pop("key_file")
            if not self._key:
                with open(key_file, "rb") as fd:
                    self._key = fd.read()
            conf["key"] = self._key

            hooks.attach(self._point, self.callable, priority=prio, **conf)
        elif policy == "dangerously_insecure":
            conf.pop("key_file", None)
            hooks.attach(self._point, authz_fake, priority=prio, **conf)
        else:
            raise Exception('unknown authz policy "%s"' % policy)



"""
Python thread pool, see
http://code.activestate.com/recipes/577187-python-thread-pool/
Author: Valentin Kuznetsov <vkuznet [AT] gmail [DOT] com>
"""
# futures
from __future__ import division

from builtins import range, object
from future import standard_library
standard_library.install_aliases()


# system modules
import time
import json
import hashlib
import threading
from queue import Queue

# WMCore modules
from Utils.PythonVersion import PY3
from WMCore.MicroService.Tools.Common import getMSLogger
from Utils.Utilities import encodeUnicodeToBytesConditional


def genkey(query):
    """
    Generate a new key-hash for a given query. We use md5 hash for the
    query and key is just hex representation of this hash.
    """
    if  isinstance(query, dict):
        record = dict(query)
        query = json.JSONEncoder(sort_keys=True).encode(record)
    keyhash = hashlib.md5()
    keyhash.update(encodeUnicodeToBytesConditional(query, condition=PY3))
    return keyhash.hexdigest()

def set_thread_name(ident, name):
    "Set thread name for given identified"
    for thr in threading.enumerate():
        if  thr.ident == ident:
            thr.name = name
            break

class StoppableThread(threading.Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self, target, name, args):
        super(StoppableThread, self).__init__(target=target, name=name, args=args)
        self._stop_event = threading.Event()

    def stop(self):
        "Set event to stop the thread"
        self._stop_event.set()

    def stopped(self):
        "Return stopped status of the thread"
        return self._stop_event.is_set()

    def running(self):
        "Return running status of the thread"
        return not self._stop_event.is_set()

def start_new_thread(name, func, args, unique=False):
    "Wrapper wroung standard thread.strart_new_thread call"
    if  unique:
        threads = sorted(threading.enumerate())
        for thr in threads:
            if  name == thr.name:
                return thr
#     thr = threading.Thread(target=func, name=name, args=args)
    thr = StoppableThread(target=func, name=name, args=args)
    thr.daemon = True
    thr.start()
    return thr

class UidSet(object):
    "UID holder keeps track of uid frequency"
    def __init__(self):
        self.set = {}

    def add(self, uid):
        "Add given uid or increment uid occurence in a set"
        if  not uid:
            return
        if  uid in self.set:
            self.set[uid] += 1
        else:
            self.set[uid] = 1

    def discard(self, uid):
        "Either discard or downgrade uid occurence in a set"
        if  uid in self.set:
            self.set[uid] -= 1
        if  uid in self.set and not self.set[uid]:
            del self.set[uid]

    def __contains__(self, uid):
        "Check if uid present in a set"
        if  uid in self.set:
            return True
        return False

    def get(self, uid):
        "Get value for given uid"
        return self.set.get(uid, 0)

class Worker(threading.Thread):
    """Thread executing worker from a given tasks queue"""
    def __init__(self, name, taskq, pidq, uidq, logger=None):
        self.logger = getMSLogger(verbose=True, logger=logger)
        threading.Thread.__init__(self, name=name)
        self.exit = 0
        self.tasks = taskq
        self.pids = pidq
        self.uids = uidq
        self.daemon = True
        self.start()

    def force_exit(self):
        """Force run loop to exit in a hard way"""
        self.exit = 1

    def run(self):
        """Run thread loop"""
        while True:
            if self.exit:
                return
            task = self.tasks.get()
            if task is None:
                return
            if self.exit:
                return
            evt, pid, func, args, kwargs = task
            try:
                func(*args, **kwargs)
                self.pids.discard(pid)
            except Exception as exc:
                self.pids.discard(pid)
                msg = "func=%s args=%s kwargs=%s" % (func, args, kwargs)
                self.logger.error('error %s, call %s', str(exc), msg)
            evt.set()

class TaskManager(object):
    """
    Task manager class based on thread module which
    executes assigned tasks concurently. It uses a
    pool of thread workers, queue of tasks and pid
    set to monitor jobs execution.

    .. doctest::

        Use case:
        mgr  = TaskManager()
        jobs = []
        jobs.append(mgr.spawn(func, args))
        mgr.joinall(jobs)

    """
    def __init__(self, nworkers=10, name='TaskManager', logger=None):
        self.logger = getMSLogger(verbose=True, logger=logger)
        self.name = name
        self.pids = set()
        self.uids = UidSet()
        self.tasks = Queue()
        self.workers = [Worker(name, self.tasks, self.pids, self.uids, logger) \
                        for _ in range(0, nworkers)]

    def status(self):
        "Return status of task manager queue"
        info = {'qsize':self.tasks.qsize(), 'full':self.tasks.full(),
                'unfinished':self.tasks.unfinished_tasks,
                'nworkers':len(self.workers)}
        return {self.name: info}

    def nworkers(self):
        """Return number of workers associated with this manager"""
        return len(self.workers)

    def spawn(self, func, *args, **kwargs):
        """Spawn new process for given function"""
        pid = kwargs.get('pid', genkey(str(args) + str(kwargs)))
        evt = threading.Event()
        if  not pid in self.pids:
            self.pids.add(pid)
            task = (evt, pid, func, args, kwargs)
            self.tasks.put(task)
        else:
            # the event was not added to task list, invoke set()
            # to pass it in wait() call, see joinall
            evt.set()
        return evt, pid

    def remove(self, pid):
        """Remove pid and associative process from the queue"""
        self.pids.discard(pid)

    def is_alive(self, pid):
        """Check worker queue if given pid of the process is still running"""
        return pid in self.pids

    def clear(self, tasks):
        """
        Clear all tasks in a queue. It allows current jobs to run, but will
        block all new requests till workers event flag is set again
        """
        _ = [t[0].clear() for t in tasks] # each task is return from spawn, i.e. a pair (evt, pid)

    def joinall(self, tasks):
        """Join all tasks in a queue and quit"""
        _ = [t[0].wait() for t in tasks] # each task is return from spawn, i.e. a pair (evt, pid)

    def quit(self):
        """Put None task to all workers and let them quit"""
        _ = [self.tasks.put(None) for _ in self.workers]
        time.sleep(1) # let workers threads cool-off and quit



#!/usr/bin/env python
"""
_PortForward_

A decorator for swapping ports in an url
"""

from builtins import str, bytes

def portForward(port):
    """
    Decorator wrapper function for port forwarding of the REST calls of any
    function to a given port.

    Currently there are three constraints for applying this decorator.
    1. The function to be decorated must be defined within a class and not being a static method.
       The reason for that is because we need to be sure the function's signature will
       always include the class instance as its first argument.
    2. The url argument must be present as the second one in the positional argument list
       of the decorated function (right after the class instance argument).
    3. The url must follow the syntax specifications in RFC 1808:
       https://tools.ietf.org/html/rfc1808.html

    If all of the above constraints are fulfilled and the url is part of the
    urlMangleList, then the url is parsed and the port is substituted with the
    one provided as an argument to the decorator's wrapper function.

    param port: The port to which the REST call should be forwarded.
    """
    def portForwardDecorator(callFunc):
        """
        The actual decorator
        """

        def portMangle(callObj, url, *args, **kwargs):
            """
            Function used to check if the url coming with the current argument list
            is to be forwarded and if so change the port to the one provided as an
            argument to the decorator wrapper.

            :param classObj: This is the class object (slef from within the class)
                             which is always to be present in the signature of a
                             public method. We will never use this argument, but
                             we need it there for not breaking the positional
                             argument order
            :param url:      This is the actual url to be (eventually) forwarded
            :param *args:    The positional argument list coming from the original function
            :param *kwargs:  The keywords argument list coming from the original function
            """
            forwarded = False
            try:
                if isinstance(url, str):
                    urlToMangle = 'https://cmsweb'
                    if url.startswith(urlToMangle):
                        newUrl = url.replace('.cern.ch/', '.cern.ch:%d/' % port, 1)
                        forwarded = True
                elif isinstance(url, bytes):
                    urlToMangle = b'https://cmsweb'
                    if url.startswith(urlToMangle):
                        newUrl = url.replace(b'.cern.ch/', b'.cern.ch:%d/' % port, 1)
                        forwarded = True

            except Exception:
                pass
            if forwarded:
                return callFunc(callObj, newUrl, *args, **kwargs)
            else:
                return callFunc(callObj, url, *args, **kwargs)
        return portMangle
    return portForwardDecorator


class PortForward():
    """
    A class with a call method implementing a simple way to use the functionality
    provided by the protForward decorator as a pure functional call:
    EXAMPLE:
        from Utils.PortForward import PortForward

        portForwarder = PortForward(8443)
        url = 'https://cmsweb-testbed.cern.ch/couchdb'
        url = portForwarder(url)
    """
    def __init__(self, port):
        """
        The init method for the PortForward call class. This one is supposed
        to simply provide an initial class instance with a logger.
        """
        self.port = port

    def __call__(self, url):
        """
        The call method for the PortForward class
        """
        def dummyCall(self, url):
            return url
        return portForward(self.port)(dummyCall)(self, url)



from __future__ import (division, print_function)
from builtins import str
from future.utils import viewitems, listvalues, listitems

import copy
import socket
from collections import defaultdict

from WMCore.Services.WMArchive.CMSSWMetrics import CMSSWMetrics

# From top level

WMARCHIVE_TOP_LEVEL_REMOVE_OUTSIDE_LAYER = ["steps"]
# convert data format under stpes["cmsRun1"/"logArch1"/"stageOut1"]["output"]
WMARCHIVE_REMOVE_OUTSIDE_LAYER = ["checksum", "dataset"]
# convert to list from str
WMARCHIVE_CONVERT_TO_LIST = ["lfn"]

WMARCHIVE_DATA_MAP = {"inputPath": "inputDataset",
                      "lfn": "outputLFNs", "input": "inputLFNs"}

WMARCHIVE_REMOVE_FIELD = ["user_dn", "user_vogroup", "user_vorole"]
WMARCHIVE_COMBINE_FIELD = {"outputDataset": ["primaryDataset", "processedDataset", "dataTier"]}

WMARCHIVE_LFN_REF_KEY = ["lfn", "files"]
WMARCHIVE_PFN_REF_KEY = []
WMARCHIVE_FILE_REF_KEY = {"LFN": WMARCHIVE_LFN_REF_KEY}

ERROR_TYPE = {'exitCode': int}

PERFORMANCE_TYPE = {'cpu': {'AvgEventCPU': float,
                            'AvgEventTime': float,
                            'MaxEventCPU': float,
                            'MaxEventTime': float,
                            'MinEventCPU': float,
                            'MinEventTime': float,
                            'TotalEventCPU': float,
                            'TotalJobCPU': float,
                            'TotalJobTime': float,
                            'EventThroughput': float,
                            'TotalLoopCPU': float,
                            'TotalInitTime': float,
                            'TotalInitCPU': float,
                            'NumberOfThreads': int,
                            'NumberOfStreams': int},
                    'memory': {'PeakValueRss': float,
                               'PeakValueVsize': float},
                    'storage': {'readAveragekB': float,
                                'readCachePercentageOps': float,
                                'readMBSec': float,
                                'readMaxMSec': float,
                                'readNumOps': float,
                                'readPercentageOps': float,
                                'readTotalMB': float,
                                'readTotalSecs': float,
                                'writeTotalMB': float,
                                'writeTotalSecs': float},
                    'cmssw': CMSSWMetrics()}

TOP_LEVEL_STEP_DEFAULT = {'analysis': {},
                          'cleanup': {},
                          'logs': {},
                          'errors': [],
                          'input': [],
                          'output': [],
                          'WMCMSSWSubprocess': {},
                          'performance': {}
                          }

# only composed value need to bs set default value
STEP_DEFAULT = {  # 'name': '',
    'analysis': {},
    'cleanup': {},
    'logs': {},
    'errors': [],
    'input': [{  # 'catalog': '',
        # 'events': -1,
        # 'guid': '',
        # 'input_source_class': '',
        # 'input_type': '',
        # 'lfn': -1,
        # 'module_label': '',
        # 'pfn': -1,
        'runs': []}],
    'output': [{  # 'acquisitionEra': '',
        # 'adler32': '',
        # 'applicationName': '',
        # 'applicationVersion': '',
        # 'async_dest': '',
        # 'branch_hash': '',
        # 'catalog': '',
        # 'cksum': '',
        # 'configURL': '',
        # 'events': -1,
        # 'globalTag': '',
        # 'guid': '',
        # 'inputDataset': '',
        'inputLFNs': [],
        # TODO change to empty string from None
        # 'location': '',
        # 'merged': False,
        # 'module_label': '',
        # 'output_module_class': '',
        # 'outputDataset': '',
        'outputLFNs': [],
        # 'prep_id': '',
        # 'processingStr': '',
        # 'processingVer': -1,
        'runs': [],
        # 'size': -1,
        # 'validStatus': '',
        # "SEName": '',
        # "PNN": '',
        # "GUID": '',
        # 'StageOutCommand': ''
        }],
    'WMCMSSWSubprocess': {},
    'performance': {'cpu': {},
                    'memory': {},
                    'multicore': {},
                    'storage': {},
                    'cmssw': {}}
    # 'site': 'T2_CH_CERN',
    # 'start': 1454569735,
    # 'status': 0,
    # 'stop': 1454569736
    }


def cleanStep(idict):
    """
    perform clean-up of PFNs attributes in place for given dictionary

    :param idict: a FWJR report dictionary
    :return: a dictionary without PFNs
    """
    for step in ['input', 'output']:
        data = idict.get(step, {})
        for key, values in data.items():
            for elem in values:
                for skip in ['pfn', 'InputPFN', 'OutputPFN', 'inputpfns']:
                    if skip in elem:
                        del elem[skip]
            data[key] = values
    return idict


def combineDataset(dataset):
    dataset["outputDataset"] = "/%s/%s/%s" % (dataset.pop("primaryDataset"),
                                              dataset.pop("processedDataset"),
                                              dataset.pop("dataTier"))
    return dataset


def changeRunStruct(runDict):
    runList = []
    for run in runDict:
        singleRun = {"runNumber": int(run)}
        singleRun.update({'lumis': [], 'eventsPerLumi': []})
        runList.append(singleRun)

    return runList


def _changeToFloat(value):
    if value in ["-nan", "nan", "inf", ""]:
        return -1.0
    return float(value)


def _validateTypeAndSetDefault(sourceDict, stepDefault):
    # check primitive time and remvoe if the values is composite type.
    for key, value in listitems(sourceDict):  # ACHTUNG! dict size changes while iterating
        if key not in stepDefault and value in [[], {}, None, "None"]:
            del sourceDict[key]

    # set missing composite type defaut.
    for category in stepDefault:
        if (category not in sourceDict) or (category in sourceDict and not sourceDict[category]):
            sourceDict[category] = stepDefault[category]


def changePerformanceStruct(perfDict):
    return [{"pName": prop, "value": _changeToFloat(value)} for prop, value in viewitems(perfDict)]


def changeToList(aDict):
    return [{"prop": prop, "value": value} for prop, value in viewitems(aDict)]


def convertInput(inputList):
    for inputDict in inputList:
        if "runs" in inputDict:
            inputDict['runs'] = changeRunStruct(inputDict["runs"])

        _validateTypeAndSetDefault(inputDict, STEP_DEFAULT['input'][0])

    return inputList


def typeCastError(errorList):
    for errorDict in errorList:
        for key in errorDict:
            if key in ERROR_TYPE:
                value = errorDict[key]
                errorDict[key] = ERROR_TYPE[key](value)
    return errorList


def typeCastPerformance(performDict):
    newPerfDict = defaultdict(dict)
    for key in PERFORMANCE_TYPE:
        if key in performDict:
            for param in PERFORMANCE_TYPE[key]:
                if key == 'cmssw':
                    # so far we skip validation of cmssw metrics values
                    # since they may differ in different CMSSW releases
                    newPerfDict[key] = performDict[key]
                    continue
                if param in performDict[key]:
                    try:
                        value = performDict[key][param]
                        if value in ["-nan", "nan", "inf", ""]:
                            value = -1
                        if PERFORMANCE_TYPE[key][param] == int:
                            # the received value comes from CMSSW FWJR and its type is string
                            # Although type is int, CMSSW FWJR string is constructed as float i.e. "3.0"
                            # In that case we convert to float first before type cast to int.
                            # since int("3.0") will raise an exception but int(float("3.0") won't
                            value = float(value)
                        newPerfDict[key][param] = PERFORMANCE_TYPE[key][param](value)
                    except ValueError as ex:
                        newPerfDict[key][param] = PERFORMANCE_TYPE[key][param](-1)
                        print("key: %s, param: %s, value: %s \n%s" % (key, param,
                                                                      performDict[key][param], str(ex)))
    return dict(newPerfDict)


def convertOutput(outputList):
    newOutputList = []
    for itemList in outputList:
        newOutputList.extend(itemList)

    for outDict in newOutputList:
        for field in WMARCHIVE_REMOVE_FIELD:
            if field in outDict:
                del outDict[field]

        for field in WMARCHIVE_CONVERT_TO_LIST:
            if field in outDict and isinstance(outDict[field], (str, bytes)):
                outDict[field] = [outDict[field]]

        for oldKey, newKey in viewitems(WMARCHIVE_DATA_MAP):
            if oldKey in outDict:
                outDict[newKey] = outDict[oldKey]
                del outDict[oldKey]

        if "runs" in outDict:
            outDict['runs'] = changeRunStruct(outDict["runs"])

        if "checksums" in outDict:
            outDict.update(outDict["checksums"])
            del outDict["checksums"]

        if "dataset" in outDict:
            outDict.update(combineDataset(outDict["dataset"]))
            del outDict["dataset"]

        if "location" in outDict and isinstance(outDict["location"], list):
            if outDict["location"]:
                outDict["location"] = outDict["location"][0]
            else:
                outDict["location"] = ""

        _validateTypeAndSetDefault(outDict, STEP_DEFAULT['output'][0])

    return newOutputList


def convertStepValue(stepValue):
    if "status" in stepValue:
        if stepValue["status"] == "Failed":
            stepValue["status"] = 1
        else:
            stepValue["status"] = int(stepValue["status"])

    if "errors" in stepValue:
        if not stepValue['errors']:
            stepValue['errors'] = []
        else:
            typeCastError(stepValue['errors'])

    input_keys = ['source', 'logArchives']
    if "input" in stepValue:
        if not stepValue['input']:
            # if empty convert to list from {}
            stepValue['input'] = []

        elif len(stepValue['input']) > 1:
            # assume only one input value
            raise Exception("more than one input value %s" % list(stepValue['input']))

        elif list(stepValue['input'].keys())[0] in input_keys:
            stepValue['input'] = convertInput(stepValue['input'][list(stepValue['input'])[0]])

        else:
            raise Exception("Unknown input key %s" % list(stepValue['input']))

    if "output" in stepValue:
        # remove output module name layer
        stepValue['output'] = convertOutput(listvalues(stepValue['output']))

    if "performance" in stepValue:
        stepValue["performance"] = typeCastPerformance(stepValue["performance"])
        # If it needs to chnage to list format replace to this
        # for category in stepValue["performance"]:
        #    stepValue["performance"][category] = changePerformanceStruct(stepValue["performance"][category])

        _validateTypeAndSetDefault(stepValue["performance"], STEP_DEFAULT["performance"])

    # If structure need to be changed with this uncomments
    # listConvKeys = ['analysis', 'cleanup', 'logs', 'parameters']
    # for key in listConvKeys:
    #    stepValue[key] = changeToList(stepValue[key])

    return stepValue


def convertSteps(steps):
    stepList = []
    for key, value in viewitems(steps):
        stepItem = {}
        stepItem['name'] = key
        value = cleanStep(value)
        stepItem.update(convertStepValue(value))
        _validateTypeAndSetDefault(stepItem, TOP_LEVEL_STEP_DEFAULT)
        stepList.append(stepItem)
    return stepList


def convertToArchiverFormat(fwjr):
    """
    """
    newFWJR = copy.deepcopy(fwjr)
    if "steps" in newFWJR:
        newFWJR["steps"] = convertSteps(newFWJR["steps"])

    return newFWJR


def createFileArrayRef(fwjr, fArrayRef):
    if isinstance(fwjr, list):
        for item in fwjr:
            createFileArrayRef(item, fArrayRef)
    elif isinstance(fwjr, dict):
        for key, value in viewitems(fwjr):
            addKeyFlag = False

            for fileType, keyList in viewitems(WMARCHIVE_FILE_REF_KEY):
                for kw in keyList:
                    if kw in key.lower():
                        fArrayRef[fileType].add(key)
                        addKeyFlag = True

            if not addKeyFlag:
                createFileArrayRef(value, fArrayRef)
    else:
        return


def createFileArray(fwjr, fArray, fArrayRef):
    if isinstance(fwjr, dict):
        for key, value in viewitems(fwjr):
            for fileType in WMARCHIVE_FILE_REF_KEY:
                if key in fArrayRef[fileType]:
                    if isinstance(value, list):
                        for fileName in value:
                            fArray[fileType].add(fileName)
                    else:  # this should be string
                        fArray[fileType].add(value)
            createFileArray(value, fArray, fArrayRef)
    elif isinstance(fwjr, list):
        for item in fwjr:
            createFileArray(item, fArray, fArrayRef)
    else:
        return


def changeToFileRef(fwjr, fArray, fArrayRef):
    if isinstance(fwjr, dict):
        for key, value in listitems(fwjr):  # ACHTUNG! dict values change while iterating
            for fileType in WMARCHIVE_FILE_REF_KEY:
                if key in fArrayRef[fileType]:
                    if isinstance(value, list):
                        newRef = []
                        for fileName in value:
                            index = fArray[fileType].index(fileName)
                            newRef.append(index)
                    else:  # this should be string
                        newRef = fArray[fileType].index(value)
                    fwjr[key] = newRef
            changeToFileRef(value, fArray, fArrayRef)
    elif isinstance(fwjr, list):
        for item in fwjr:
            changeToFileRef(item, fArray, fArrayRef)
    else:
        return


def createArchiverDoc(job, version=None):
    """
    job_id is jobid + retry count same as couch db _id
    """

    job_id = job["id"]
    fwjr = job['doc']["fwjr"]
    jobtype = job['doc']["jobtype"]
    jobstate = job['doc']['jobstate']
    create_ts = job['doc']['timestamp']
    newfwjr = convertToArchiverFormat(fwjr)

    fArrayRef = {}
    fArray = {}
    for fileType in WMARCHIVE_FILE_REF_KEY:
        fArrayRef[fileType] = set()
        fArray[fileType] = set()

    createFileArrayRef(newfwjr, fArrayRef)

    for fileType in WMARCHIVE_FILE_REF_KEY:
        fArrayRef[fileType] = list(fArrayRef[fileType])

    createFileArray(newfwjr, fArray, fArrayRef)

    for fileType in WMARCHIVE_FILE_REF_KEY:
        fArray[fileType] = list(fArray[fileType])

    changeToFileRef(newfwjr, fArray, fArrayRef)

    # convert to fwjr format

    for fileType in WMARCHIVE_FILE_REF_KEY:
        newfwjr["%sArrayRef" % fileType] = fArrayRef[fileType]
        newfwjr["%sArray" % fileType] = fArray[fileType]

    if version is None:
        # add this trry to remove the dependency on WMCore code.
        import WMCore
        version = WMCore.__version__
    # append meta data in fwjr
    wnName = ""
    if "WorkerNodeInfo" in fwjr:
        wnName = fwjr["WorkerNodeInfo"].get("HostName", "")

    newfwjr['meta_data'] = {'agent_ver': version,
                            'host': socket.gethostname().lower(),
                            'wn_name': wnName,
                            'fwjr_id': job_id,
                            'jobtype': jobtype,
                            'jobstate': jobstate,
                            'ts': create_ts
                            }
    return newfwjr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Module based on the pycurl_manager module, implementing a few
functions to call to the Rucio RESTful APIs, leveraging the
pycurl concurrency.
"""
from __future__ import print_function, division, absolute_import

from builtins import str
from future.utils import viewitems

from future import standard_library
standard_library.install_aliases()

import datetime
import json
import logging
import re

from urllib.parse import quote, unquote

from Utils.CertTools import cert, ckey
from WMCore.Services.pycurl_manager import RequestHandler
from WMCore.Services.pycurl_manager import getdata as multi_getdata

### Amount of days that we wait for stuck rules to be sorted
### After that, the rule is not considered and a new rule is created
STUCK_LIMIT = 7  # 7 days


def parseNewLineJson(stream):
    """
    Parse newline delimited json streaming data
    """
    for line in stream.split("\n"):
        if line:
            yield json.loads(line)


def stringDateToEpoch(strDate):
    """
    Given a date/time in the format of:
        'Thu, 29 Apr 2021 13:15:42 UTC'
    it returns an integer with the equivalent EPOCH time
    :param strDate: a string with the date and time
    :return: the equivalent EPOCH time (integer)
    """
    timestamp = datetime.datetime.strptime(strDate, "%a, %d %b %Y %H:%M:%S %Z")
    return int(timestamp.strftime('%s'))


def getRucioToken(rucioAuthUrl, rucioAcct):
    """
    Provided a Rucio account, fetch a token from the authentication server
    :param rucioAuthUrl: url to the rucio authentication server
    :param rucioAcct: rucio account to be used
    :return: an integer with the expiration time in EPOCH
    """
    params = {}
    headers = {"X-Rucio-Account": rucioAcct}

    url = '%s/auth/x509' % rucioAuthUrl
    logging.info("Requesting a token to Rucio for account: %s, against url: %s", rucioAcct, rucioAuthUrl)
    mgr = RequestHandler()
    res = mgr.getheader(url, params=params, headers=headers, ckey=ckey(), cert=cert())
    if res.getReason() == "OK":
        userToken = res.getHeaderKey('X-Rucio-Auth-Token')
        tokenExpiration = res.getHeaderKey('X-Rucio-Auth-Token-Expires')
        logging.info("Retrieved Rucio token valid until: %s", tokenExpiration)
        # convert the human readable expiration time to EPOCH time
        tokenExpiration = stringDateToEpoch(tokenExpiration)
        return userToken, tokenExpiration

    raise RuntimeError("Failed to acquire a Rucio token. Error: {}".format(res.getReason()))


def renewRucioToken(rucioAuthUrl, userToken):
    """
    Provided a user Rucio token, check it's lifetime and extend it by another hour
    :param rucioAuthUrl: url to the rucio authentication server
    :param rucioAcct: rucio account to be used
    :return: a datetime.datetime object with the new token lifetime
    """
    params = {}
    headers = {"X-Rucio-Auth-Token": userToken}

    url = '%s/auth/validate' % rucioAuthUrl
    logging.info("Renewing the Rucio token...")
    mgr = RequestHandler()
    res = mgr.getdata(url, params=params, headers=headers, ckey=ckey(), cert=cert())
    try:
        newExpiration = eval(res)['lifetime']
    except Exception as exc:
        raise RuntimeError("Failed to renew Rucio token. Response: {} Error: {}".format(res, str(exc)))
    return newExpiration


def getPileupContainerSizesRucio(containers, rucioUrl, rucioToken, scope="cms"):
    """
    Given a list of containers, find their total size in Rucio
    :param containers: list of container names
    :param rucioUrl: a string with the Rucio URL
    :param rucioToken: a string with the user rucio token
    :param scope: a string with the Rucio scope of our data
    :return: a flat dictionary of container and their respective sizes
    NOTE: Value `None` is returned in case the data-service failed to serve a given request.
    NOTE: Rucio version of getPileupDatasetSizes()
    """
    sizeByDset = {}
    if not containers:
        return sizeByDset

    headers = {"X-Rucio-Auth-Token": rucioToken}

    urls = ['{}/dids/{}/{}?dynamic=anything'.format(rucioUrl, scope, cont) for cont in containers]
    logging.info("Executing %d requests against Rucio for the container size", len(urls))
    data = multi_getdata(urls, ckey(), cert(), headers=headers)

    for row in data:
        container = row['url'].split('/dids/{}/'.format(scope))[1]
        container = container.replace("?dynamic=anything", "")
        if row['data'] is None:
            msg = "Failure in getPileupContainerSizesRucio for container {}. Response: {}".format(container, row)
            logging.error(msg)
            sizeByDset.setdefault(container, None)
            continue
        response = json.loads(row['data'])
        try:
            sizeByDset.setdefault(container, response['bytes'])
        except KeyError:
            msg = "getPileupContainerSizesRucio function did not return a valid response for container: %s. Error: %s"
            logging.error(msg, container, response)
            sizeByDset.setdefault(container, None)
            continue
    return sizeByDset


def listReplicationRules(containers, rucioAccount, grouping,
                         rucioUrl, rucioToken, scope="cms"):
    """
    List all the replication rules for the input filters provided.
    It builds a dictionary of container name and the locations where
    they have a rule locking data on, with some additional rule state
    logic in the code.
    :param containers: list of container names
    :param rucioAccount: string with the rucio account
    :param grouping: rule grouping string, only "A" or "D" are allowed
    :param rucioUrl: string with the Rucio url
    :param rucioToken: string with the Rucio token
    :param scope: string with the data scope
    :return: a flat dictionary key'ed by the container name, with a list of RSE
      expressions that still need to be resolved
    NOTE: Value `None` is returned in case the data-service failed to serve a given request.
    NOTE-2: Available rule states can be found at:
    https://github.com/rucio/rucio/blob/16f39dffa1608caa0a1af8bbc0fcff2965dccc50/lib/rucio/db/sqla/constants.py#L180
    """
    locationByContainer = {}
    if not containers:
        return locationByContainer
    if grouping not in ["A", "D"]:
        raise RuntimeError("Replication rule grouping value provided ({}) is not allowed!".format(grouping))

    headers = {"X-Rucio-Auth-Token": rucioToken}
    urls = []
    for cont in containers:
        urls.append('{}/rules/?scope={}&account={}&grouping={}&name={}'.format(rucioUrl, scope, rucioAccount,
                                                                               grouping, quote(cont, safe="")))
    logging.info("Executing %d requests against Rucio to list replication rules", len(urls))
    data = multi_getdata(urls, ckey(), cert(), headers=headers)

    for row in data:
        container = unquote(row['url'].split("name=")[1])
        if "200 OK" not in row['headers']:
            msg = "Failure in listReplicationRules for container {}. Response: {}".format(container, row)
            logging.error(msg)
            locationByContainer.setdefault(container, None)
            continue
        try:
            locationByContainer.setdefault(container, [])
            for item in parseNewLineJson(row['data']):
                if item['state'] in ["U", "SUSPENDED", "R", "REPLICATING", "I", "INJECT"]:
                    msg = "Container %s has a rule ID %s in state %s. Will try to create a new rule."
                    logging.warning(msg, container, item['id'], item['state'])
                    continue
                elif item['state'] in ["S", "STUCK"]:
                    if item['error'] == 'NO_SOURCES:NO_SOURCES':
                        msg = "Container {} has a STUCK rule with NO_SOURCES.".format(container)
                        msg += " Data could be lost forever... Rule info is: {}".format(item)
                        logging.warning(msg)
                        continue

                    # then calculate for how long it's been stuck
                    utcTimeNow = int(datetime.datetime.utcnow().strftime('%s'))
                    if item['stuck_at']:
                        stuckAt = stringDateToEpoch(item['stuck_at'])
                    else:
                        # consider it to be stuck since its creation
                        stuckAt = stringDateToEpoch(item['created_at'])

                    daysStuck = (utcTimeNow - stuckAt) // (24 * 60 * 60)
                    if daysStuck > STUCK_LIMIT:
                        msg = "Container {} has a STUCK rule for {} days (limit set to: {}).".format(container,
                                                                                                     daysStuck,
                                                                                                     STUCK_LIMIT)
                        msg += " Not going to use it! Rule info: {}".format(item)
                        logging.warning(msg)
                        continue
                    else:
                        msg = "Container {} has a STUCK rule for only {} days.".format(container, daysStuck)
                        msg += " Considering it for the pileup location"
                        logging.info(msg)
                else:
                    logging.info("Container %s has rule ID %s in state %s, using it.",
                                 container, item['id'], item['state'])

                ### NOTE: this is not an RSE name, but an RSE expression that still needs to be resolved
                locationByContainer[container].append(item['rse_expression'])
        except Exception as exc:
            msg = "listReplicationRules function did not return a valid response for container: %s."
            msg += "Server responded with: %s\nError: %s"
            logging.exception(msg, container, str(exc), row['data'])
            locationByContainer.setdefault(container, None)
            continue
    return locationByContainer


def getPileupSubscriptionsRucio(datasets, rucioUrl, rucioToken, scope="cms"):
    """
    Provided a list of datasets, find dataset level subscriptions where it's
    as complete as `percent_min`.
    :param datasets: list of dataset names
    :param rucioUrl: a string with the Rucio URL
    :param rucioToken: a string with the user rucio token
    :param scope: a string with the Rucio scope of our data
    :return: a dictionary of datasets and a list of their location.
    NOTE: Value `None` is returned in case the data-service failed to serve a given request.
    """
    # FIXME: we should definitely make a feature request to Rucio...
    # so much, just to get the final RSEs for a container!!!
    locationByDset = {}
    if not datasets:
        return locationByDset

    headers = {"X-Rucio-Auth-Token": rucioToken}
    # first, resolve the dataset into blocks
    blocksByDset = getContainerBlocksRucio(datasets, rucioUrl, rucioToken, scope)
    urls = []
    for _dset, blocks in viewitems(blocksByDset):
        if blocks:
            for block in blocks:
                urls.append('{}/replicas/{}/{}/datasets'.format(rucioUrl, scope, quote(block)))

    # this is going to be bloody expensive in terms of HTTP requests
    logging.info("Executing %d requests against Rucio replicas API for blocks", len(urls))
    data = multi_getdata(urls, ckey(), cert(), headers=headers)
    for row in data:
        block = row['url'].split("/{}/".format(scope))[1]
        block = unquote(re.sub("/datasets$", "", block, 1))
        container = block.split("#")[0]
        locationByDset.setdefault(container, set())
        if row['data'] is None:
            msg = "Failure in getPileupSubscriptionsRucio container {} and block {}.".format(container, block)
            msg += " Response: {}".format(row)
            logging.error(msg)

            locationByDset[container] = None
            continue
        if locationByDset[container] is None:
            # then one of the block requests failed, skip the whole dataset
            continue
        thisBlockRSEs = set()
        for item in parseNewLineJson(row['data']):
            if item['state'] == "AVAILABLE":
                thisBlockRSEs.add(item["rse"])
        logging.info("Block: %s is available at: %s", block, thisBlockRSEs)
        # now we have the final block location
        if not locationByDset[container]:
            # then this is the first block of this dataset
            locationByDset[container] = thisBlockRSEs
        else:
            # otherwise, make an intersection of them
            locationByDset[container] = locationByDset[container] & thisBlockRSEs
    return locationByDset


def getBlocksAndSizeRucio(containers, rucioUrl, rucioToken, scope="cms"):
    """
    Given a list of containers, find all their correspondent blocks and their sizes.
    :param containers: list of container names
    :param rucioUrl: a string with the Rucio URL
    :param rucioToken: a string with the user rucio token
    :param scope: a string with the Rucio scope of our data
    :return: a dictionary in the form of:
    {"dataset":
        {"block":
            {"blockSize": 111, "locations": ["x", "y"]}
        }
    }
    NOTE: Value `None` is returned in case the data-service failed to serve a given request.
    NOTE2: meant to return an output similar to Common.getBlockReplicasAndSize
    """
    contBlockSize = {}
    if not containers:
        return contBlockSize

    headers = {"X-Rucio-Auth-Token": rucioToken}
    urls = []
    for cont in containers:
        urls.append('{}/dids/{}/dids/search?type=dataset&long=True&name={}'.format(rucioUrl, scope, quote(cont + "#*")))
    logging.info("Executing %d requests against Rucio DIDs search API for containers", len(urls))
    data = multi_getdata(urls, ckey(), cert(), headers=headers)
    for row in data:
        container = row['url'].split("name=")[1]
        container = unquote(container).replace("#*", "")
        contBlockSize.setdefault(container, {})
        if row['data'] in [None, ""]:
            msg = "Failure in getBlocksAndSizeRucio function for container {}. Response: {}".format(container, row)
            logging.error(msg)
            contBlockSize[container] = None
            continue

        for item in parseNewLineJson(row['data']):
            # NOTE: we do not care about primary block location in Rucio
            contBlockSize[container][item['name']] = {"blockSize": item['bytes'], "locations": []}
    return contBlockSize


### NOTE: likely not going to be used for a while
def getContainerBlocksRucio(containers, rucioUrl, rucioToken, scope="cms"):
    """
    Provided a list of containers, find all their blocks.
    :param containers: list of container names
    :param rucioUrl: a string with the Rucio URL
    :param rucioToken: a string with the user rucio token
    :param scope: a string with the Rucio scope of our data
    :return: a dictionary key'ed by the datasets with a list of blocks.
    NOTE: Value `None` is returned in case the data-service failed to serve a given request.
    """
    blocksByDset = {}
    if not containers:
        return blocksByDset

    headers = {"X-Rucio-Auth-Token": rucioToken}
    urls = ['{}/dids/{}/{}/dids'.format(rucioUrl, scope, cont) for cont in containers]
    logging.info("Executing %d requests against Rucio DIDs API for blocks in containers", len(urls))
    data = multi_getdata(urls, ckey(), cert(), headers=headers)
    for row in data:
        container = row['url'].split("/{}/".format(scope))[1]
        container = re.sub("/dids$", "", container, 1)
        if not row['data']:
            logging.warning("Dataset: %s has no blocks in Rucio", container)
        blocksByDset.setdefault(container, [])
        for item in parseNewLineJson(row['data']):
            blocksByDset[container].append(item["name"])
    return blocksByDset


### NOTE: likely not going to be used for a while
def getBlockReplicasAndSizeRucio(datasets, rucioUrl, rucioToken, scope="cms"):
    """
    Given a list of datasets, find all their blocks with replicas
    available.
    :param datasets: list of dataset names
    :param rucioUrl: a string with the Rucio URL
    :param rucioToken: a string with the user rucio token
    :param scope: a string with the Rucio scope of our data
    :return: a dictionary in the form of:
    {"dataset":
        {"block":
            {"blockSize": 111, "locations": ["x", "y"]}
        }
    }
    NOTE: Value `None` is returned in case the data-service failed to serve a given request.
    """
    dsetBlockSize = {}
    if not datasets:
        return dsetBlockSize

    headers = {"X-Rucio-Auth-Token": rucioToken}
    # first, figure out their block names
    blocksByDset = getContainerBlocksRucio(datasets, rucioUrl, rucioToken, scope=scope)
    urls = []
    for _dset, blocks in viewitems(blocksByDset):
        for block in blocks:
            urls.append('{}/replicas/{}/{}/datasets'.format(rucioUrl, scope, quote(block)))

    # next, query the replicas API for the block location
    # this is going to be bloody expensive in terms of HTTP requests
    logging.info("Executing %d requests against Rucio replicas API for blocks", len(urls))
    data = multi_getdata(urls, ckey(), cert(), headers=headers)
    for row in data:
        block = row['url'].split("/{}/".format(scope))[1]
        block = unquote(re.sub("/datasets$", "", block, 1))
        container = block.split("#")[0]
        dsetBlockSize.setdefault(container, dict())
        if row['data'] is None:
            msg = "Failure in getBlockReplicasAndSizeRucio for container {} and block {}.".format(container, block)
            msg += " Response: {}".format(row)
            logging.error(msg)

            dsetBlockSize[container] = None
            continue
        if dsetBlockSize[container] is None:
            # then one of the block requests failed, skip the whole dataset
            continue

        thisBlockRSEs = []
        blockBytes = 0
        for item in parseNewLineJson(row['data']):
            blockBytes = item['bytes']
            if item['state'] == "AVAILABLE":
                thisBlockRSEs.append(item["rse"])
        # now we have the final block location
        if not blockBytes and not thisBlockRSEs:
            logging.warning("Block: %s has no replicas and no size", block)
        else:
            dsetBlockSize[container][block] = {"locations": thisBlockRSEs, "blockSize": blockBytes}
    return dsetBlockSize
